"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

import os, sys, stat
import copy
import curses
from math import sqrt, pow

import graph
from collection import *
from selector import Selector
from config import *
from workloads import *

_TEMPLATE_CACHE={}
def _lookup_template(filename):
    if not _TEMPLATE_CACHE.has_key(filename):
        _TEMPLATE_CACHE[filename] = Template(filename=filename)
    return _TEMPLATE_CACHE[filename]

def _vector_add(a, b):
    assert a or b
    if a == None:
        return b
    if b == None:
        return a
    assert isinstance(a, (list, tuple))
    assert isinstance(b, (list, tuple))
    assert len(a) == len(b)
    ret = []
    for i in range(len(a)):
        ret.append(a[i] + b[i])
    return ret

def html_fmt_kernel(kernel):
    return '<span class="kernel">%s</span>' % (get_kernel_display(kernel),)

def html_fmt_kernels(kernels):
    ret = []
    for k in kernels:
        ret.append(html_fmt_kernel(k))
    return ret

def html_fmt_nfsopt(nfsopt):
    return '<span class="nfsopt">%s</span>' % (get_nfsopt_display(nfsopt),)

def html_fmt_group(g):
    assert isinstance(g, Selector)
    return '<div class="group">%s%s</div>' % \
            (html_fmt_nfsopt(g.nfsopt), html_fmt_kernel(g.kernel))

def make_report_path(report_type, collection, selection):
    path = 'Report_' + str(report_type)

    sdisplay = selection.display(collection.selection)

    if sdisplay:
        path += '_' + '_'.join(sdisplay)

    # erase certain chars
    for c in ['/', ':']:
        path = path.replace(c, '')

    return path.replace(' ', '_') + '.html'

def make_report_title(report_type, collection, selection):
    title = "%s" % (report_type,)

    sdisplay = selection.display(collection.selection)

    if sdisplay:
        title += ': ' + ', '.join(sdisplay)

    return title

# TODO there must be a better way to do this!
def vals_empty(groups, keys, vals, report, ignore_keys=None):
    if not ignore_keys:
        ignore_keys = set()

    def _is_empty_none_only(x):
        if x != None:
            return False
        return True

    def _is_empty_none_or_zero(x):
        if x:
            return False
        return True

    is_empty = _is_empty_none_or_zero

    if report.show_zeros:
        is_empty = _is_empty_none_only

    for g in groups:
        if not vals.has_key(g):
            continue

        for k in keys:
            if k in ignore_keys:
                continue

            if not vals[g].has_key(k):
                continue

            v = vals[g][k]
            if isinstance(v, (list, tuple)):
                for v2 in v:
                    if not is_empty(v2):
                        return False
            else:
                if not is_empty(v):
                    return False
    return True

def stat_info_id(sel, statbin_name, key):
    return repr((sel, statbin_name, key)).replace(',', '+').replace("'", '')

def stat_info_actions(sel, statbin_name, key):
    r = [ "info_select(this, '%s')" % stat_info_id(subsel, statbin_name, key)
          for subsel in sel.foreach() ]
    return 'info_start(this);' + '; '.join(r) + ' ; info_mode(this);'

class StatusOutput:
    def __init__(self, f):
        self.f = f
        self.temp = 0

        w = curses.initscr()
        height, width = w.getmaxyx()
        del w
        curses.endwin()
        self.maxwidth = width

    def erase_temporary(self):
        if self.temp:
            self.f.write('\r' + ' ' * self.temp + '\r')
            self.temp = 0

    def clear(self):
        self.erase_temporary()

    def write(self, msg, temporary=False):
        if msg and len(msg) > self.maxwidth:
            msg = msg[:self.maxwidth]

        self.erase_temporary()

        self.f.write(msg)
        self.f.flush()

        if temporary:
            self.temp += len(msg)

STATUS_OUTPUT = StatusOutput(sys.stdout)

class BucketDef:
    """
        Used to define buckets - accumulation is done by BucketHits()
    """
    def __init__(self, bucket2keys):
        """
            bucket2keys - a map of bucket name (str) -> list of keys (str)
        """
        self.bucket2keys = bucket2keys
        self.key2bucket = {}
        # make map of key -> bucket
        for b, ks in bucket2keys.iteritems():
            for k in ks:
                self.key2bucket[k] = b

    def find_suffix(self, search, suffixes):
        """
            Split 'search' into (name, suffix)

            suffixes - list of suffixes
        """
        for s in suffixes:
            if search.endswith('_' + s):
                idx = len(search) - len('_' + s)
                return (search[:idx], search[idx+1:])
        raise KeyError("key has invaid suffix: %r" % (search))

    def has_suffix(self, search, suffixes):
        try:
            self.find_suffix(search, suffixes)
        except:
            return False
        return True

    def key_to_bucket(self, key, other_name, suffixes):
        k, s = self.find_suffix(key, suffixes)
        return '%s_%s' % (self.key2bucket.get(k, other_name), s)

    def bucket_to_keys(self, bucket, suffixes):
        b, s = self.find_suffix(bucket, suffixes)
        return '%s_%s' % (self.bucket2keys[b], s)

    def buckets(self, other_name, suffixes):
        """ return all buckets """
        r = []
        for s in suffixes:
            r.extend([ '%s_%s' % (x, s) for x in self.bucket2keys.keys()])
        r.sort()
        for s in suffixes:
            r.append('%s_%s' % (other_name, s))
        return r

    def bucket_info(self, statbin_name, selection, report, bucket_hits,
                    groups, keys, vals, other_name, suffixes):
        descmap = {}
        bettermap = {}
        for b in self.buckets(other_name, suffixes):
            links = []
            better = None
            for bk in bucket_hits.bucket_keys(b):
                bucket, s_ = self.find_suffix(bk, suffixes)
                title = self.find_suffix(bk, suffixes)[0]
                bkidx = bucket_hits.get_hatch_idx(b, bk)
                links.append('%s %s' % (get_hatch_html(bkidx, title), bucket))

                # lookup better for this key
                keybetter = report.collection.stat_better(statbin_name, bk)
                if not better:
                    better = keybetter
                else:
                    assert better == keybetter

            descmap[b] = '<br>'.join(links)
            bettermap[b] = better or BETTER_UNKNOWN

        return descmap, bettermap


nfsstat_bucket_def = BucketDef({
  'creation and deletion':
    ('create',
     'open',
     'open_conf',
     'open_dgrd',
     'mkdir',
     'rmdir',
     'remove',
     'close',
     'mknod',
    ),

  'file metadata operations':
    ('access',
     'lookup',
     'lookup_root',
     'rename',
     'link',
     'readlink',
     'symlink',
    ),

  'readdir operations':
    ('readdir',
     'readdirplus',
    ),

  'attr operations':
    ('getattr',
     'setattr',
    ),

  'fs metadata operations':
    ('fsstat',
     'fsinfo',
     'statfs',
    ),

  'locking operations':
    ('lock',
     'lockt',
     'locku',
     'rel_lkowner',
     'delegreturn',
     'get_lease_t',
    ),

  'write operations':
    ('write',
     'commit',
     'ds_write',
    ),

  'read operations':
    ('read',
    ),

  'pnfs operations':
    ('getdevinfo',
     'getdevlist',
     'layoutget',
     'layoutcommit',
     'layoutreturn',
    ),

  'acl operations':
    ('getacl',
     'setacl',
    ),

  'session operations':
    ('create_ses',
     'destroy_ses',
     'exchange_id',
    ),
})

mountstat_bucket_def = BucketDef({
  'creation and deletion':
    ('CREATE',
     'OPEN',
     'MKDIR',
     'RMDIR',
     'REMOVE',
     'CLOSE',
     'OPEN_CONFIRM',
     'OPEN_DOWNGRADE',
    ),

  'file metadata operations':
    ('ACCESS',
     'LOOKUP',
     'LOOKUP_ROOT',
     'RENAME',
     'LINK',
     'READLINK',
     'SYMLINK',
    ),

  'readdir operations':
    ('READDIR',
     'READDIRPLUS',
    ),

  'attr operations':
    ('GETATTR',
     'SETATTR',
    ),

  'fs metadata operations':
    ('FSSTAT',
     'FSINFO',
     'STATFS',
    ),

  'locking operations':
    ('LOCK',
     'LOCKT',
     'LOCKU',
     'RELEASE_LOCKOWNER',
     'DELEGRETURN',
     'GET_LEASE_TIME',
    ),

  'write operations':
    ('WRITE',
     'COMMIT',
    ),

  'read operations':
    ('READ',
    ),

  'pnfs operations':
    ('GETDEVICEINFO',
     'GETDEVICELIST',
     'LAYOUTGET',
     'LAYOUTCOMMIT',
     'LAYOUTRETURN',
    ),

  'acl operations':
    ('GETACL',
     'SETACL',
    ),

  'session operations':
    ('CREATE_SESSION',
     'DESTROY_SESSION',
     'EXCHANGE_ID',
    ),
})

class BucketHits:
    def __init__(self):
        self._hits = {}
        self.total = 0.0
        self._bucketkey2hatch = {}
        self._bucket2keys_ordered = {}

    def add(self, g, b, key, val):
        assert not self._bucketkey2hatch
        assert not self._bucket2keys_ordered

        def _val(v):
            if isinstance(v, (list, tuple)):
                return np.mean([ float(x) for x in v ])
            return float(v)

        if not self._hits.has_key(g):
            self._hits[g] = {}
        if not self._hits[g].has_key(b):
            self._hits[g][b] = {}
        if not self._hits[g][b].has_key(key):
            self._hits[g][b][key] = 0.0

        val = _val(val)
        self._hits[g][b][key] += float(val)
        self.total += float(val)

    def has_hit(self, g, b, k):
        return self._hits.get(g, {}).get(b, {}).has_key(k)

    def get_val(self, g, b, k):
        return self._hits.get(g, {}).get(b, {}).get(k, None)

    def hits(self, g, b):
        """ return a list unique hit keys for a given (g,b) """
        self._make_mapping(b)
        hset = set(self._hits[g].get(b, {}).keys())
        return [ h for h in self._bucket2keys_ordered[b] if h in hset ]

    def empty(self, b):
        for g in self._hits.keys():
            for h in self._hits[g].get(b, {}).keys():
                if self._hits[g][b][h]:
                    return False
        return True

    def bucket_keys(self, b):
        self._make_mapping(b)
        return self._bucket2keys_ordered[b]

    def _make_mapping(self, b):
        if self._bucketkey2hatch.has_key(b):
            return

        # calc magnitude of each key across ALL groups for ordering
        key2val = {}
        for g in self._hits.keys():
            for k in self._hits[g].get(b, {}).keys():
                if not key2val.has_key(k):
                    key2val[k] = 0.0
                key2val[k] += self._hits[g][b][k]

        key2val = [ (k, v) for k, v in key2val.iteritems() ]
        key2val.sort(lambda x,y: cmp(x[1], y[1]))
        key2val.reverse()

        k2h = {}
        ordered = []
        for i, kv in enumerate(key2val):
            k = kv[0]
            assert not k2h.has_key(k)
            k2h[k] = i
            ordered.append(k)

        self._bucketkey2hatch[b] = k2h
        self._bucket2keys_ordered[b] = ordered

    def get_hatch_idx(self, bucket, k):
        self._make_mapping(bucket)
        return self._bucketkey2hatch[bucket][k]


class Table:
    def __init__(self, values, groups, keys, units,
                 nolegend=False,
                 fmt_key=None,
                 fmt_group=None,
                 fmt_cell=None,
                 wide_data=False):
        assert isinstance(values, dict)

        self.values = {}

        for group in groups:
            self.values[group] = {}
            for k in keys:
                self.values[group][k] = values[group].get(k, None)

        self.groups = groups
        self.keys = keys
        self.units = units

        self.nolegend = nolegend
        self.fmt_key = fmt_key
        self.fmt_cell = fmt_cell
        self.fmt_group = fmt_group

        self._classes = ['data',]

        if wide_data:
            self._classes.append('wide_data')

    def classes(self):
        return ' '.join(self._classes)

    def html_key(self, k):
        """ return key formatted for html """
        if self.fmt_key:
            return self.fmt_key(k)
        return k

    def html_group(self, g):
        """ return group formatted for html """
        def _calc_group_idx():
            group_idx = None
            for i, findg in enumerate(self.groups):
                if g == findg:
                    group_idx = i
                    break
            return group_idx

        # add legend
        legend = ''
        if not self.nolegend:
            group_idx = _calc_group_idx()
            if group_idx != None:
                legend = '<div class="legend_%u"></div>' % \
                    (color_idx(group_idx),)

        # format group text
        if self.fmt_group:
            g = self.fmt_group(g)

        return '<div class="cell">%s</div><div class="cell">%s</div>' % \
                (legend, str(g))

    def html_cell(self, g, k):
        cell = None
        classes = []

        if self.values.has_key(g) and \
           self.values[g].has_key(k) and \
           self.values[g][k] != None:
            val = self.values[g][k]

            if isinstance(val, graph.Stack):
                comb = None
                for stack_item in val:
                    comb = _vector_add(comb, [ float(x) for x in stack_item['value']])
                valmean = np.mean(comb)
                valstd = np.std(comb)

                cell = '%.02f' % valmean
                fmt_std = '%.02f' % valstd
                if fmt_std != '0.00':
                    cell += ' <span class="stddev">%s%s</span>' % \
                                (PLUSMINUS, fmt_std)


            elif isinstance(val, (list, tuple)):
                val = [ float(x) for x in val ]
                valmean = np.mean(val)
                valstd = np.std(val)

                cell = '%.02f' % valmean
                fmt_std = '%.02f' % valstd
                if fmt_std != '0.00':
                    cell += ' <span class="stddev">%s%s</span>' % \
                                (PLUSMINUS, fmt_std)

            else:
                cell = str(val)

            if self.units:
                cell += ' <span class="units">%s</span>' % (self.units,)

        if cell == None:
            cell = NO_DATA_HTML
        if not cell:
            cell = COMPARISON_ZERO_HTML

        if cell in (NO_DATA_HTML, COMPARISON_ZERO_HTML):
            classes.append('no_data')
        else:
            cell = """<table class="celltable"><tr>
                       <td> %s </td>
                      </tr></table>""" % cell

        if self.fmt_cell:
            cell = self.fmt_cell(g, k, self.values, cell)

        if classes:
            classes = ' '.join(classes)
            classes = ' class="%s"' % classes
        else:
            classes = ''

        return '<td%s>%s</td>' % (classes, cell)

    def html(self):
        template = _lookup_template(TEMPLATE_TABLE)
        return template.render(table=self)

class TocNode:
    def __init__(self, text, section, parent):
        self.text = text
        self.section = section
        self.parent = parent
        self.children = []

    def title_list(self):
        tl = []
        node = self
        while node:
            if node.text:
                tl.insert(0, node.text)
            node = node.parent
        return tl

    def title(self):
        tl = self.title_list()
        return ' : '.join(tl)

    def anchor(self):
        return '<a name="%s"></a>' % self.section

    def num(self):
        return '<span class="section_num">%s</span>' % self.section

    def add(self, text):
        if self.section:
            section = '%s.%s' % (self.section, len(self.children) + 1)
        else:
            section = '%s' % (len(self.children) + 1,)

        new_node = TocNode(text, section, self)
        self.children.append(new_node)

        return new_node

    def html(self):
        if not self.text:
            template = _lookup_template(TEMPLATE_TOC)
        else:
            template = _lookup_template(TEMPLATE_TOCNODE)
        return template.render(node=self)

class Dataset:
    def __init__(self, selection, widget, title, statbin_name, units,
                 groups, keys, vals, data_info, toc, report,
                 no_graph=False, no_title=False,
                 key_desc=None, bettermap=None,
                 tall_cell=False, subtitle='', graph_ignore_keys=None,
                 anchor=None, fmt_key=None, fmt_cell=None, fmt_group=None):
        """ generate the "dataset" - a graph and a table  """
        self.toc = toc.add(title)
        self.wide_data = False

        no_ylabel = False
        if report.is_comparison:
            tablevals = copy.deepcopy(vals)
            self.make_comparison_vals(tablevals, keys, selection)
            no_ylabel = True
            units = ''
        else:
            tablevals = vals

        # Graph section
        self.graph_tag = ''
        if not no_graph:
            if not graph_ignore_keys:
                graph_ignore_keys = []
            graph_keys = [ k for k in keys if not k in graph_ignore_keys ]
            self.wide_data = (len(graph_keys) * len(groups)) > WIDE_DATA_THRESHOLD

            uniq_tuple = (report.title, widget, selection,
                          statbin_name)
            self.graph_tag = report.graphs.make_graph('bar', uniq_tuple,
                    units, vals, groups, graph_keys, no_ylabel, tall_cell)

        if statbin_name:
            def link_fmt(key, chain=None):
                key_display = key
                if chain:
                    key_display = chain(key)

                binfo = report.collection.get_better_info(selection,
                                statbin_name, k, bettermap=bettermap)
                if binfo:
                    binfo = '<span class="better_sym">%s</span> ' % binfo[0]
                else:
                    binfo = ''

                actions = stat_info_actions(selection, statbin_name, key)

                return binfo + \
                    """<span class="statlink" onclick="%s">%s</span>""" % \
                      (actions, key_display)

            old = fmt_key
            fmt_key = lambda x: link_fmt(x, chain=old)

        def _make_cell_data_info(g, k):
            info = data_info[g][k]
            template = _lookup_template(TEMPLATE_CELLDATAINFO)
            return template.render(info=info)

        infos = []
        for g in groups:
            for k in keys:
                infos.append(_make_cell_data_info(g, k))
        self.infos_html = '\n'.join(infos)

        tab = Table(tablevals, groups, keys, units,
                    fmt_key=fmt_key, fmt_cell=fmt_cell,
                    fmt_group=fmt_group,
                    wide_data=self.wide_data)
        self.table_html = tab.html()

        self.keys = keys
        self.subtitle = subtitle
        self.statbin_name = statbin_name
        self.anchor = anchor
        self.key_desc = key_desc
        self.selection = selection
        self.bettermap = bettermap
        self.report = report
        if fmt_key:
            self.fmt_key = fmt_key
        else:
            self.fmt_key = lambda x: x


        # no more TOC children

    # TODO clean this up!
    def html_key_info(self, k):
        description = \
            self.report.collection.stat_description(self.statbin_name, k,
                                                    descmap=self.key_desc)
        if not description:
            return ''

        binfo = self.report.collection.get_better_info(self.selection,
                                self.statbin_name, k, bettermap=self.bettermap)
        if binfo:
            binfo = """
             <span class="better_sym">%s</span> <span class="better">(%s)</span>
            """ % (binfo[0], binfo[1])
        else:
            binfo = ''

        template = _lookup_template(TEMPLATE_DATASETKEYINFO)
        return template.render(key_display=self.fmt_key(k),
                               description=description,
                               binfo=binfo)


    def html(self):
        template = _lookup_template(TEMPLATE_DATASET)
        return template.render(dataset=self)

    def make_comparison_vals(self, vals, keys, selection):
        def _get_mean_std(val):
            if isinstance(val, graph.Stack):
                comb = None
                for stack_item in val:
                    comb = _vector_add(comb, [ float(x) for x in stack_item['value']])
                return (np.mean(comb), np.std(comb))
            elif isinstance(val, (float, long, int)):
                return (val, 0.0)
            elif isinstance(val, (tuple, list)):
                _val = [ float(v) for v in val ]
                return (np.mean(_val), np.std(_val))
            else:
                raise ValueError('Unexpected type: %r' % val)

        for nfsopts_idx, nsel in enumerate(selection.foreach_nfsopt()):
            for key in keys:
                collect = []

                for ksel in nsel.foreach_kernel():
                    collect.append((ksel, vals[ksel].get(key, None)))

                # find the first entry with *some* data to use as reference
                ref_idx = None

                for i in range(0, len(collect)):
                    ref_comp_key = collect[i][0]
                    if vals[ref_comp_key].get(key, None) == None:
                        continue
                    ref_val, ref_std = _get_mean_std(vals[ref_comp_key][key])

                    if ref_val or ref_std:
                        ref_idx = i
                        group_idx = (nfsopts_idx * len(selection.kernels)) + i
                        refstr = '<div class="legend_small_%u"></div>' % \
                                    (color_idx(group_idx),)
                        break

                for i in range(0, len(collect)):
                    c = collect[i]

                    # handle no data
                    if vals[c[0]].get(key, None) == None:
                        vals[c[0]][key] = NO_DATA_HTML
                        continue

                    v, v_std = _get_mean_std(vals[c[0]][key])

                    # handle no data
                    if not v and not v_std:
                        vals[c[0]][key] = COMPARISON_ZERO_HTML
                        continue

                    # if something has data there better be a reference point
                    assert ref_idx != None

                    # handle reference point
                    if i == ref_idx:
                        tmp = """
                            <table class="celltable"><tr>
                             <td> %s </td>
                            </tr></table>
                        """ % (refstr)

                        vals[c[0]][key] = tmp
                        continue

                    # handle default: has data, not reference point
                    def pct_f(x, y):
                        if y == 0.0:
                            raise Exception('Zero Division')
                            return 0.0
                        pct = (float(x) / float(y)) * 100.0
                        return pct

                    diff_val = pct_f(v - ref_val, ref_val)
                    stddev = sqrt((pow(ref_std,2) + pow(v_std,2)) / 2.0)
                    diff_std = pct_f(stddev, ref_val)

                    operator = '-'
                    if diff_val >= 0.0:
                       operator = '+'

                    diff_val_str = '%0.2f' % abs(diff_val)
                    diff_std_str = '%0.2f' % abs(diff_std)

                    if diff_val_str == '0.00' and diff_std_str == '0.00':
                        operator = '='
                        vstr = '<td><b>%s</b></td><td>%s</td>' % \
                            (operator, refstr)
                    else:
                        stddevstr = ''
                        if diff_std_str != '0.00':
                            stddevstr = ' <span class="stddev">%s%s</span>' % \
                                (PLUSMINUS, diff_std_str)
                        vstr = '<td>%s</td><td><b>%s</b> %s%s%%</td>' % \
                            (refstr, operator, diff_val_str, stddevstr)

                    vstr = '<table class="celltable"><tr>%s</tr></table>' % vstr
                    vals[c[0]][key] = vstr

    def make_heading(self, workload, widget_name, dataset_name):
        return ' - '.join((workload, widget_name, dataset_name))

#
# WIDGETS
#
class Widget:
    def __init__(self, collection, selection, report, toc):
        assert self.widget
        assert self.desc
        assert self.statbin_name

        self.collection = collection
        self.selection = selection
        self.report = report

        self.toc = toc.add(self.widget)
        self.datasets = []

        note_mesgs = [ f(selection) for f in getattr(self, 'notes', []) ]
        self.note_mesgs = [ m for m in note_mesgs if m ]

        self.setup()

    def setup(self):
        raise NotImplemented

    def gather_data(self, keys):
        groups = []
        vals = {}

        self.collection.set_used_statbin_keys(self.statbin_name, keys)

        for subsel in self.selection.foreach():
            sbin = self.collection.get_testbin(subsel).get_statbin(self.statbin_name)
            assert not vals.has_key(subsel)
            vals[subsel] = {}

            for k in keys:
                vals[subsel][k] = sbin.get(k, None)

            groups.append(subsel)

        info = self.make_data_info(self.collection, groups, keys)

        return groups, vals, info


    def gather_buckets(self, keys):
        self.collection.set_used_statbin_keys(self.statbin_name, keys)

        groups = []
        vals = {}
        bucket_hits = BucketHits()

        for subsel in self.selection.foreach():
            sbin = self.collection.get_testbin(subsel).get_statbin(self.statbin_name)

            assert not vals.has_key(subsel)
            vals[subsel] = {}

            # add all values together for each bucket
            # this adds per-run since value arrays are in run load order
            for k in keys:
                try:
                    new = sbin.get(k)
                except KeyError:
                    continue

                b = self.bucket_def.key_to_bucket(k, self.other_name, self.suffixes)

                bucket_hits.add(subsel, b, k, new)

                if not vals[subsel].has_key(b):
                    vals[subsel][b] = graph.Stack()
                else:
                    assert len(vals[subsel][b][0]['value']) == len(new)
                vals[subsel][b].append({'key': k,
                                       'value': tuple(new)})

            groups.append(subsel)

        info = self.make_data_info(self.collection, groups,
                self.bucket_def.buckets(self.other_name, self.suffixes),
                bucket_hits=bucket_hits)

        # post-process hatch index
        for g in vals.keys():
            for b in vals[g].keys():
                stack = vals[g][b]
                for stack_item in stack:
                    stack_item['hidx'] = bucket_hits.get_hatch_idx(b,
                                                        stack_item['key'])
                # sort stack based on hidx
                stack.sort(lambda x,y: cmp(x['hidx'], y['hidx']))

        return groups, vals, info, bucket_hits

    def make_data_info(self, collection, groups, keys,
                       bucket_hits=None):
        data_info = {}

        for group_idx, g in enumerate(groups):
            data_info[g] = {}

            def stat_lookup(x):
                res = []
                for subsel in g.foreach():
                    testbin = collection.get_testbin(subsel)
                    sbin = testbin.get_statbin(self.statbin_name)

                    v = sbin.get(x, None)
                    if v:
                        assert len(testbin.workloaddirs) == len(v)
                        res.append((v, testbin.workloaddirs, sbin.filename))

                if res:
                    # ???
                    assert len(res) == 1
                    return res[0]

                return None, None, None

            for k in keys:
                if bucket_hits:
                    name = "%s::bucket('%s')" % (self.statbin_name, k)
                else:
                    name = '%s::%s' % (self.statbin_name, k)

                values = {}
                num_runs = []
                valkeys = []
                workloaddirs = None
                filename = None

                if bucket_hits:
                    for h in bucket_hits.hits(g, k):
                        v, dirs, fname = stat_lookup(h)

                        if v:
                            values[h] = v
                            num_runs.append(len(v))
                            valkeys.append(h)
                            if not workloaddirs:
                                workloaddirs = dirs
                            else:
                                assert workloaddirs == dirs
                            if not filename:
                                filename = fname
                            else:
                                assert filename == fname
                else:
                    v, dirs, fname = stat_lookup(k)
                    if v:
                        values[name] = v
                        num_runs.append(len(v))
                        valkeys.append(name)
                        workloaddirs = dirs
                        filename = fname

                if values:
                    assert min(num_runs) == max(num_runs)
                    num_runs = max(num_runs)
                    assert len(workloaddirs) == num_runs
                else:
                    num_runs = 0

                stat_id = stat_info_id(g, self.statbin_name, k)

                info = {'name':      name,
                        'selection': g,
                        'keys':      valkeys,
                        'num_runs':  num_runs,
                        'values':    values,
                        'workloaddirs':  workloaddirs,
                        'file':      filename,
                        'key':       k,
                        'id':        stat_id,
                        'color_idx': color_idx(group_idx),
                        'is_bucket': bucket_hits != None,
                       }
                data_info[g][k] = info

        return data_info

    def new_dataset(self, selection, title, groups, keys, vals, data_info,
                    **kwargs):
        ignore_keys=kwargs.get('graph_ignore_keys', set())
        # don't add empty datasets
        if vals_empty(groups, keys, vals, self.report, ignore_keys=ignore_keys):
            return None

        collection = self.report.collection

        units = ''
        for k in keys:
            if k in ignore_keys:
                continue
            this = collection.stat_units(self.statbin_name, k)
            if not units:
                units = this
            else:
                assert units == this, \
                    'unit mismatch! statbin: %s, keys: %s' % \
                        (self.statbin_name, keys)

        kwargs['graph_ignore_keys'] = ignore_keys
        if not 'fmt_group' in kwargs:
            kwargs['fmt_group'] = html_fmt_group

        new = Dataset(selection, self.widget, title, self.statbin_name,
                      units, groups, keys, vals, data_info,
                      self.toc, self.report,
                      **kwargs)
        self.datasets.append(new)
        return new

    def html(self):
        template = _lookup_template(TEMPLATE_WIDGET)
        return template.render(widget=self, dataset_class=Dataset)


class SimpleWidget(Widget):
    def __init__(self, collection, selection, report, toc):
        assert len(self.ds_info)

        self.keys = set()
        for t, keys in self.ds_info:
            for k in keys:
                if k in self.keys:
                    raise ValueError("key %s already used in ds_info" % k)
                self.keys.add(k)

        self.keys = tuple(self.keys)

        Widget.__init__(self, collection, selection, report, toc)

    def setup(self):
        groups, vals, data_info = self.gather_data(self.keys)
        for t, keys in self.ds_info:
            self.new_dataset(self.selection, t, groups, keys, vals, data_info)

class Widget_RunTimes(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = 'Run Times'
        self.desc =   'Run times of test from time(1)'
        self.notes = [note_filebench_times]
        self.statbin_name =   'times'

        self.ds_info = (
            ('Real Time',          ('time_real',)),
            ('Sys and User Times', ('time_sys', 'time_user')),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)


class Widget_Filebench(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "Filebench Output"
        self.desc =  "Output from the filebench test suite"
        self.statbin_name =  'filebench'


        self.ds_info = (
            ('Operation count',   ('op_count',)),
            ('Operations/second', ('ops_per_second',)),
            ('MB/second',         ('mb_per_second',)),
            ('CPU/Operation',     ('cpu_per_op',)),
            ('Latency',           ('latency_ms',)),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_NfsBytes(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "Byte Counts"
        self.desc =  """ A count from mountstats(1) of bytes read and written by
                         syscalls and NFS operations.
                     """
        self.statbin_name =  'mountstats'

        self.ds_info = (
            ('Read and Write Syscalls',
                    ('read_normal', 'write_normal')),
            ('Read and Write Syscalls (O_DIRECT)',
                    ('read_odirect', 'write_odirect')),
            ('Read and Write NFS calls',
                    ('read_nfs', 'write_nfs')),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_RpcCounts(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "RPC Counts"
        self.desc =  """ RPC message counts from nfsstats(1).  """
        self.statbin_name =  'nfsstats'

        self.ds_info = (
            ('Calls', ('rpc_calls',)),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_RpcStats(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "RPC Statistics"
        self.desc =  """ RPC message counts from mountstats(1).  """
        self.notes = [note_v3_no_lock, note_v41_pnfs_no_ds]
        self.statbin_name =  'mountstats'

        self.ds_info = (
            ('Requests and Replies', ('rpc_requests', 'rpc_replies')),
        )

        # TODO: Add to sanity check like if xid_not_found is really big
        #'xid_not_found',
        #'backlog_queue_avg',

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_Nfsiostat(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = 'NFS Throughput'
        self.desc =  'Statistics gathered from nfsiostat(1)'
        self.notes = [note_v41_pnfs_no_ds]
        self.statbin_name =  'nfsiostat'

        self.ds_info = (
            ('Read/Write KB/s',
                ('read_kb_per_sec', 'write_kb_per_sec')),
            ('Read/Write operations/s',
                ('read_ops_per_sec', 'write_ops_per_sec')),
            ('Read/Write Average KB per Operation',
                ('read_kb_per_op', 'write_kb_per_op')),
            ('Read/Write Average RTT',
                ('read_avg_rtt_ms', 'write_avg_rtt_ms')),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_VfsEvents(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = 'VFS Events'
        self.desc =  'Counters for events from the VFS layer'
        self.statbin_name =  'proc_mountstats'

        self.ds_info = (
            ('Open / Lookup Counts',
                ('vfs_open', 'vfs_lookup')),
            ('Access / Updatepage Counts',
                ('vfs_access', 'vfs_updatepage')),
            ('Readpage / Readpages Counts',
                ('vfs_readpage', 'vfs_readpages')),
            ('Writepage / Writepages Counts',
                ('vfs_writepage', 'vfs_writepages')),
            ('Getdents / Setattr Counts',
                ('vfs_getdents', 'vfs_setattr')),
            ('Flush / Fsync Counts',
                ('vfs_flush', 'vfs_fsync')),
            ('Lock / Release Counts',
                ('vfs_lock', 'vfs_release')),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_InvalidateEvents(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = 'Validation Events'
        self.desc =  'Counters for invalidate events'
        self.statbin_name =  'proc_mountstats'

        self.ds_info = (
            ('Inode / Dentry Revalidate',
                ('inode_revalidate', 'dentry_revalidate')),
            ('Data / Attr Invalidate',
                ('data_invalidate', 'attr_invalidate')),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_PnfsReadWrite(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = 'pNFS Read/Write counts'
        # XXX counts or bytes??
        self.desc =  'Counters for pNFS reads and writes'
        self.statbin_name =  'proc_mountstats'

        self.ds_info = (
            ('pNFS Read/Write counts',
                ('pnfs_read', 'pnfs_write')),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)

class Widget_OtherEvents(SimpleWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = 'Other Events'
        self.desc =  'Counters for other NFS events'
        self.statbin_name =  'proc_mountstats'

        self.ds_info = (
            ('Short Read/Write Counts',
                ('short_read', 'short_write')),
            ('Congestion Wait Count',
                ('congestion_wait',)),
            ('Extend Write Count',
                ('extend_write',)),
            ('Setattr Truncate Count',
                ('setattr_trunc',)),
            ('Delay Count',
                ('delay',)),
            ('Silly Rename Counts',
                ('silly_rename',)),
        )

        SimpleWidget.__init__(self, collection, selection, report, toc)



class BucketWidget(Widget):
    suffixes = None
    bucket_def = None
    other_name = 'other'

    def __init__(self, collection, selection, report, toc):
        assert self.suffixes and len(self.suffixes)
        assert self.bucket_def
        assert self.other_name

        Widget.__init__(self, collection, selection, report, toc)

    def setup(self):
        keys = [ x 
            for x in self.collection.get_valid_statbin_keys(self.statbin_name)
            if self.bucket_def.has_suffix(x, self.suffixes) ]

        groups, vals, data_info, bucket_hits = self.gather_buckets(keys)

        descmap, bettermap = \
            self.bucket_def.bucket_info(self.statbin_name, self.selection,
                                        self.report,
                                        bucket_hits, groups, keys, vals,
                                        self.other_name, self.suffixes)

        def fmt_cell_hits(g, b, values, cellstr):
            hits = bucket_hits.hits(g, b)
            if hits:
                out = []
                for k in hits:
                    k_display = self.bucket_def.find_suffix(k, self.suffixes)[0]
                    k_val = bucket_hits.get_val(g, b, k)
                    k_hidx = bucket_hits.get_hatch_idx(b, k)
                    over = "%s: %s" % (k_display, k_val)
                    out.append(get_hatch_html(k_hidx, over))
                cellstr += '<div class="cellhits">%s</div>' % ' '.join(out)
            return cellstr

        def fmt_key(x):
            return self.bucket_def.find_suffix(x, self.suffixes)[1]

        def fmt_title(x):
            return self.bucket_def.find_suffix(x, self.suffixes)[0]

        buckets = self.bucket_def.buckets(self.other_name, self.suffixes)
        for i in range(0, len(buckets), len(self.suffixes)):
            if all([ bucket_hits.empty(b)
                     for b in buckets[i:i+len(self.suffixes)] ]):
                continue

            first = self.bucket_def.find_suffix(buckets[i], self.suffixes)
            for check_idx in range(1, len(self.suffixes)):
                this = self.bucket_def.find_suffix(buckets[i + check_idx],
                                                   self.suffixes)
                assert first[0] == this[0], \
                    "first (%r) != this (%r)" % (first, this)
                assert first[1] != this[1], \
                    "first (%r) has same suffix as this (%r)" % \
                        (first, this)

            self.new_dataset(self.selection, fmt_title(buckets[i]), groups,
                             buckets[i:i + len(self.suffixes)], vals, data_info,
                             key_desc=descmap,
                             bettermap=bettermap,
                             fmt_key=fmt_key,
                             fmt_cell=fmt_cell_hits,
                             tall_cell=True)

class Widget_NfsOpsCount(BucketWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "NFS Operation counts"
        self.desc = """ Counts from nfsstats(1) of every NFS operation
                        broken down by type.
                    """
        self.notes = [note_v3_no_lock]
        self.statbin_name = 'nfsstats'

        self.suffixes = ('count',)
        self.bucket_def = nfsstat_bucket_def

        BucketWidget.__init__(self, collection, selection, report, toc)


class Widget_NfsOpsExecRtt(BucketWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "Exec and RTT by Operation Group"
        self.desc = """ Counts from mountstats(1) of execution time and RTT
                        for each Operation.
                    """
        self.notes = [note_v3_no_lock, note_v41_pnfs_no_ds]
        self.statbin_name = 'mountstats'

        self.suffixes = ('exectime', 'rtt')
        self.bucket_def = mountstat_bucket_def

        BucketWidget.__init__(self, collection, selection, report, toc)

class Widget_NfsOpsBytes(BucketWidget):
    def __init__(self, collection, selection, report, toc):
        self.widget = "Avg Bytes by Operation Group"
        self.desc = """ Counts from mountstats(1)
                    """
        self.notes = [note_v3_no_lock, note_v41_pnfs_no_ds]
        self.statbin_name = 'mountstats'

        self.suffixes = ('avg_bytes_sent', 'avg_bytes_received')
        self.bucket_def = mountstat_bucket_def

        BucketWidget.__init__(self, collection, selection, report, toc)

class Section:
    def __init__(self, collection, selection, report, toc):
        self.collection = collection

        self.toc = toc.add(selection.workload)

        num_runs_list = self.collection.num_runs_list(kernel=selection.kernels,
                                                 workload=selection.workload)
        data_attrs = self.collection.attrs(kernel=selection.kernels,
                                           workload=selection.workload)

        if min(num_runs_list) == max(num_runs_list):
            data_attrs['num_runs'] = str(max(num_runs_list))
        else:
            data_attrs['num_runs'] = '%s-%s' % \
                    (min(num_runs_list), max(num_runs_list))
        data_attrs['command'] = workload_command(selection.workload)
        data_attrs['description'] = workload_description(selection.workload)

        data_attrs['kernels'] = html_fmt_kernels(selection.kernels)

        data_attr_ks = data_attrs.keys()
        data_attr_ks.sort()

        data_attr_ks.remove('command')
        data_attr_ks.remove('description')
        data_attr_ks.insert(0, 'command')
        data_attr_ks.insert(1, 'description')

        self.data_attrs = data_attrs
        self.data_attr_keys = data_attr_ks

        self.widgets = []

        for widget_class in report.widget_classes:
            filebench_classes = (Widget_Filebench,)
            if widget_class in filebench_classes and \
               not selection.workload.startswith('filebench_'):
                continue

            #STATUS_OUTPUT.write('%s: %s' % (selection.workload,
                                             #widget_class.__name__),
                                #temporary=True)
            w = widget_class(self.collection, selection, report, self.toc)
            self.widgets.append(w)

        #STATUS_OUTPUT.clear()

    def html(self):
        template = _lookup_template(TEMPLATE_SECTION)
        return template.render(section=self)

class Report:
    report_type = None
    show_zeros = False
    widget_classes = []

    def __init__(self, rptset, collection, selection,
                 reportdir, graphs, cssfile, is_comparison):
        self.rptset = rptset
        self.collection = collection
        self.selection = selection

        self.reportdir = reportdir
        self.graphs = graphs
        self.cssfile = cssfile

        self.is_comparison = is_comparison

        assert self.report_type

        self.toc = TocNode(None, None, None)
        self.title = make_report_title(self.report_type, collection, selection)

        #print "Generating report: %s" % self.title

        # XXX this is an array because it used to support multiple tests
        #     maybe get rid of it?
        self.sections = []

        if not self.validate_report_data(selection):
            print
            print "Skipping report %s: invalid report data" % self.title
        else:
            s = Section(collection, selection, self, self.toc)
            self.sections.append(s)

    def validate_report_data(self, selection):
        valid = True
        for test_key in selection.foreach():
            if not self.collection.has_test(test_key):
                valid = False
                warn("missing test: %r" % (test_key,))
                # don't break so warnings are generated for all missing data

        if not valid:
            print >>sys.stderr, \
                "Skipping report(s) for kernel(s) '%s', workload '%s'" % \
                    (', '.join([ get_kernel_display(x)
                               for x in selection.kernels]), selection.workload)
            print >>sys.stderr, " Ignored directories:"

            for testkey in selection.foreach():
                if self.collection.has_test(testkey):
                    for t in self.collection.get_testbin(testkey).workloaddirs:
                        print >>sys.stderr, '    %s' % t

        return valid

    def html(self):
        template = _lookup_template(TEMPLATE_REPORT)
        return template.render(report=self)

    def write_file(self, selection):
        path = make_report_path(self.report_type, self.collection, selection)
        abs_path = os.path.join(self.reportdir, path)
        file(abs_path, 'w+').write(self.html())
        print "Saved: %s" % path
        sys.stdout.flush()

class Report_Averages(Report):
    report_type = 'Averages' # poor name choice
    widget_classes = [
               Widget_Filebench,
               Widget_RunTimes,
               Widget_NfsBytes,
               Widget_Nfsiostat,
               Widget_RpcCounts,
               Widget_NfsOpsCount,
               Widget_NfsOpsExecRtt,
               Widget_NfsOpsBytes,
               Widget_VfsEvents,
               Widget_InvalidateEvents,
               Widget_PnfsReadWrite,
               Widget_OtherEvents,
              ]

class Report_Comparison(Report):
    report_type = 'Comparison'
    widget_classes = [
               Widget_Filebench,
               Widget_RunTimes,
               Widget_NfsBytes,
               Widget_Nfsiostat,
               Widget_RpcCounts,
               Widget_NfsOpsCount,
               Widget_NfsOpsExecRtt,
               Widget_NfsOpsBytes,
               Widget_VfsEvents,
               Widget_InvalidateEvents,
               Widget_PnfsReadWrite,
               Widget_OtherEvents,
              ]

class ReportSet:
    toplevel_reports = [ Report_Comparison ]
    nfsopt_reports =   [ Report_Comparison ]
    kernel_reports =   [ Report_Averages ]

    def __init__(self, collection, serial_graph_gen):
        self.collection = collection
        self.reportdir = collection.resultsdir

        imagedir = os.path.join(self.reportdir, 'images')
        self.graphs = graph.GraphFactory(imagedir,
                                   serial_gen=serial_graph_gen,
                                   status_output=STATUS_OUTPUT)

        self.cssfilepath = CSSFILEPATH
        self.cssfile = os.path.split(self.cssfilepath)[-1]

        self.jqueryurl = JQUERY_URL

        self.jsfilepath = JSFILEPATH
        self.jsfile = os.path.split(self.jsfilepath)[-1]

        self._clear_files()

        self._write_extrafiles()
        self.report_index = []
        self.report_index_last_workload = None
        self.generate_report_index()

    def _clear_files(self):
        os.system("rm %s/*.html 2> /dev/null" % self.reportdir)
        os.system("rm %s/*.css 2> /dev/null" % self.reportdir)

    # add legend to styles:
    def _write_extrafiles(self):
        os.system('cp %s %s' % (self.jsfilepath, self.reportdir))
        os.system('cp %s %s' % (self.cssfilepath, self.reportdir))
        f = file(os.path.join(self.reportdir, self.cssfile), 'a')

        for i, color in enumerate(COLORS):
            f.write("""div.legend_%u {
                        background-color: %s;
                        height: 15px;
                        width: 15px;
                        margin: 2px;
                        border: 1px solid #101010;
                    }
                    """ % (i, color))

            f.write("""div.legend_small_%u {
                        background-color: %s;
                        height: 10px;
                        width: 10px;
                        /* float: left; */
                        margin: 2px;
                        border: 1px solid #101010;
                        border-radius: 10px;
                    }
                    """ % (i, color))

    def _step_through_reports(self, cb_f):
        def _apply(report_classes, xsel, is_comparison=False):
            for report_class in report_classes:
                cb_f(report_class, xsel, is_comparison=is_comparison)

        for wsel in self.collection.selection.foreach_workload():
            if len(wsel.kernels) > 1:
                # make comparison reports
                _apply(self.toplevel_reports, wsel, is_comparison=True)

                if len(wsel.nfsopts) > 1:
                    for nsel in wsel.foreach_nfs_version():
                        _apply(self.nfsopt_reports, nsel, is_comparison=True)

            for ksel in wsel.foreach_kernel():
                _apply(self.kernel_reports, ksel)
    
    def index_add(self, report_class, selection, is_comparison=False):
        path = make_report_path(report_class.report_type, self.collection,
                                selection)
        title = make_report_title(report_class.report_type, self.collection,
                                  selection)

        # insert test divider
        if self.report_index_last_workload and \
           self.report_index_last_workload != selection.workload:
            self.report_index.append(('', ''))
        self.report_index_last_workload = selection.workload

        self.report_index.append((path, title))

    def generate_report_index(self):
        self._step_through_reports(self.index_add)

    def generate_report(self, report_class, selection, is_comparison=False):
        r = report_class(self, self.collection, selection,
                         self.reportdir, self.graphs,
                         self.cssfile, is_comparison)
        r.write_file(selection)

    def generate_reports(self):
        print
        self.write_file()
        self._step_through_reports(self.generate_report)
        print
        self.graphs.wait_for_graphs()

    def html_file(self):
        """ generate an index file """
        template = _lookup_template(TEMPLATE_INDEX)
        return template.render(index=self)

    def html(self, current_title=''):
        """ generate a <div> tag with the index in it """
        template = _lookup_template(TEMPLATE_REPORTLIST)
        return template.render(index=self, current_title=current_title)

    def html_load_info(self):
        info = self.collection.load_info()
        template = _lookup_template(TEMPLATE_LOADINFO)
        return template.render(info=info)

    def write_file(self):
        path = 'index.html'
        abs_path = os.path.join(self.reportdir, path)
        file(abs_path, 'w+').write(self.html_file())
        print "Saved: %s" % path
        sys.stdout.flush()


