"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

from config import *

class Selector:
    """ This class is used to specify selection of the current query
    """
    def __init__(self, workloads, kernels, nfsopts):
        if not isinstance(workloads, (list, tuple)):
            workloads = [workloads]
        workloads = workloads

        if not isinstance(kernels, (list, tuple)):
            kernels = [kernels]
        kernels = kernels

        if not isinstance(nfsopts, (list, tuple)):
            nfsopts = [nfsopts]
        nfsopts = nfsopts

        # make tuples to ensure const
        self.workloads = tuple(workloads)
        self.kernels = tuple(kernels)
        self.nfsopts = tuple(nfsopts)

    def __str__(self):
        return "workload%s: %s, kernel%s: %s, nfsopt%s: %s" % \
            (pluralize(len(self.workloads)), ', '.join(self.workloads),
             pluralize(len(self.kernels)), ', '.join(self.kernels),
             pluralize(len(self.nfsopts)), ', '.join(self.nfsopts))

    def __hash__(self):
        return hash((self.workloads, self.kernels, self.nfsopts))

    def html(self):
        return """workload%s: %s<br>
                  kernel%s: %s<br>
                  nfsopt%s: %s""" % \
            (pluralize(len(self.workloads)),
                ', '.join(self.workloads),
             pluralize(len(self.kernels)),
                ', '.join([ get_kernel_display(k) for k in self.kernels ]),
             pluralize(len(self.nfsopts)),
                ', '.join([ get_nfsopt_display(n) for n in self.nfsopts ]))

    def __cmp__(self, other):
        r = cmp(self.workloads, other.workloads)
        if r:
            return r
        r = cmp(self.kernels, other.kernels)
        if r:
            return r
        return cmp(self.nfsopts, other.nfsopts)

    def __repr__(self):
        return 'Selector(%r,%r,%r)' % (self.workloads, self.kernels, self.nfsopts)

    def _get_one_workload(self):
        assert len(self.workloads) == 1, \
            "%r: Expected one workload!" % self
        return self.workloads[0]
    workload = property(_get_one_workload)

    def _get_one_kernel(self):
        assert len(self.kernels) == 1, \
            "%r: Expected one kernel!" % self
        return self.kernels[0]
    kernel = property(_get_one_kernel)

    def _get_one_nfsopt(self):
        assert len(self.nfsopts) == 1, \
            "%r: Expected one nfsopt!" % self
        return self.nfsopts[0]
    nfsopt = property(_get_one_nfsopt)

    def foreach(self):
        for tsel in self.foreach_workload():
            for nsel in tsel.foreach_nfsopt():
                for ksel in nsel.foreach_kernel():
                    yield ksel

    def foreach_workload(self):
        for workload in self.workloads:
            yield Selector(workload, self.kernels, self.nfsopts)

    def foreach_nfsopt(self):
        for nfsopt in self.nfsopts:
            yield Selector(self.workload, self.kernels, nfsopt)

    def foreach_nfs_version(self):
        nfsopt_map = {}
        versions = []

        for nfsopt in self.nfsopts:
            mountopt, tags = nfsopt_to_mountopts_tags(nfsopt)
            if not mountopt in versions:
                versions.append(mountopt)
            if not nfsopt_map.has_key(mountopt):
                nfsopt_map[mountopt] = []
            nfsopt_map[mountopt].append(nfsopt)

        for vers in versions:
            yield Selector(self.workload, self.kernels, nfsopt_map[vers])

    def has_mountopt(self, match):
        for nfsopt in self.nfsopts:
            mountopt, tags = nfsopt_to_mountopts_tags(nfsopt)

            if mountopt == match:
                return True
        return False

    def foreach_kernel(self):
        for kernel in self.kernels:
            yield Selector(self.workload, kernel, self.nfsopts)

    def make_data_group(self, subsel):
        assert len(subsel.workloads) == 1
        nfsopts_display = get_nfsopt_display(subsel.nfsopts)
        kernel_display = get_kernel_display(subsel.kernel)

        if len(self.kernels) == 1:
            return nfsopts_display
        return (kernel_display, nfsopts_display)

    def display(self, all_selector):
        display = []

        if self.workloads != all_selector.workloads:
            # all or one expected
            display.append('workload %s' % get_workload_display(self.workload))

        if self.kernels != all_selector.kernels:
            # all or one expected
            display.append('kernel %s' % get_kernel_display(self.kernel))

        if self.nfsopts != all_selector.nfsopts:
            if len(self.nfsopts) == 1:
                display.append('nfsopt %s' % get_nfsopt_display(self.nfsopt))
            else:
                versions = {}
                for nfsopt in self.nfsopts:
                    mountopt, tags = nfsopt_to_mountopts_tags(nfsopt)
                    versions[mountopt] = True
                assert len(versions) == 1

                display.append('mountopt %s' % get_nfsopt_display(mountopt))

        return display
