"""
Copyright 2012 NetApp, Inc. All Rights Reserved,
contribution by Weston Andros Adamson <dros@netapp.com>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
"""

import sys
import os
import subprocess
import errno
import time
import re

from cmd import *
from config import *
from workloads import *
import settings

_server_path_v4 = re.compile('([^:]+):(\S+)')
_server_path_v6 = re.compile('(\[\S+\]):(\S+)')

class TraceError(Exception):
    pass

class TraceNotFound(TraceError):
    pass

class TraceRunning(TraceError):
    pass

class TraceAttrs:
    def __init__(self, filename=None, temp=False, new=False):
        self.__attrs = {}
        if not filename:
            self.__attrfile = os.path.join(RUNNING_TRACE_DIR, TRACE_ATTRFILE)
        else:
            self.__attrfile = filename
        self.__temp = temp

        if not temp and not new:
            try:
                f = file(self.__attrfile)
            except IOError, e:
                raise TraceNotFound('No active NFS trace found')
            for line in f:
                if line.strip():
                    name, val = line.strip().split('=',1)
                    self.__attrs[name.strip()] = val.strip()

        if not self.__attrs.has_key('tracedir'):
            self.__attrs['tracedir'] = RUNNING_TRACE_DIR

        if not self.__attrs.has_key('stoptime'):
            self.__attrs['stoptime'] = 'ongoing'

        if not self.__attrs.has_key('tracedir_version'):
            if new or temp:
                self.__attrs['tracedir_version'] = TRACE_DIR_VERSION
            else:
                self.__attrs['tracedir_version'] = 1

        self._upgrade_attrs()

    def _upgrade_attrs(self):
        """ based on the tracedir_version, upgrade attrs """
        vers = int(self.__attrs['tracedir_version'])

        if vers == TRACE_DIR_VERSION:
            return
        assert vers < TRACE_DIR_VERSION

        #print "upgrading tracedir from version %r to %r" % \
        #    (vers, TRACE_DIR_VERSION)

        while vers < TRACE_DIR_VERSION:
            if vers == 1:
                # move tags from separate attrs to one 'tags' attr
                v1_tags = [ 'delegations_enabled', 'pnfs_enabled', 'remote' ]
                tag_names = [ x for x in self.__attrs.keys() if x in v1_tags ]

                for name in tag_names:
                    assert int(self.__attrs[name]) == 1
                    del self.__attrs[name]

                self.__attrs['tags'] = ','.join(tag_names)

            elif vers == 2:
                # 'test' attr -> 'workload' attr 
                self.__attrs['workload'] = self.__attrs['test']
                del self.__attrs['test']

            else:
                raise Exception("Unhandled tracedir_version: %s" % (vers,))

            vers += 1

        assert vers == TRACE_DIR_VERSION
        self.__attrs['tracedir_version'] = TRACE_DIR_VERSION

    def _sorted_names(self):
        names = self.__attrs.keys()
        names.sort()
        return names

    def get(self, name, *args):
        if self.__attrs.has_key(name):
            return self.__attrs[name]

        # handle optional default value
        if args:
            assert len(args) == 1
            return args[0]

        raise KeyError(name)

    def set(self, name, value):
        self.__attrs[name] = value
        self._write()

    def to_dict(self):
        return self.__attrs

    def __str__(self):
        o = []
        o.append('TraceAttrs:')
        for k in self._sorted_names():
            o.append('  %-10s: %s' % (k, self.__attrs[k]))
        return '\n'.join(o)

    def _write(self):
        if self.__temp:
            return
        f = file(self.__attrfile, 'w+')
        for k, v in self.__attrs.iteritems():
            f.write('%s = %s\n' % (k, v))

def _dir_create():
    #print 'creating directory: %s' % RUNNING_TRACE_DIR
    try:
        os.mkdir(RUNNING_TRACE_DIR)
    except OSError, e:
        if e.errno == errno.EEXIST:
            raise TraceRunning('An NFS trace is already running')
        raise

def dir_remove_old_asides():
    res = cmd('ls %s-* 2> /dev/null' % (RUNNING_TRACE_DIR,), raiseerrorcode=False)

    if '\n'.join(res[0]).strip():
        inform("Removing old error result directories from %s-*" %
                (RUNNING_TRACE_DIR,))
        cmd('rm -rf %s-* 2> /dev/null' % (RUNNING_TRACE_DIR,))
        print

def _dir_aside():
    #print 'removing directory: %s' % RUNNING_TRACE_DIR
    newdir = '%s-error-%u' % (RUNNING_TRACE_DIR, int(time.time()))
    warn('Moving failed rundir to %s - it will be deleted on next run of this script!' % newdir)
    cmd('mv "%s" "%s"' % (RUNNING_TRACE_DIR, newdir))


def _mount(attrs):
    cmdstr = 'sudo mount -v -t nfs -o "%s" "%s" "%s"' % \
                (attrs.get('mountopts'),
                 attrs.get('serverpath'),
                 attrs.get('localpath'))
    cmd(cmdstr, raiseerrorout=False)

def _try_mount(attrs):
    sys.stdout.write("Mounting: %s..." % attrs.get('serverpath'))
    sys.stdout.flush()

    err = None
    for i in range(TRACE_MOUNT_TRIES):
        try:
            _mount(attrs)
        except Exception, e:
            sys.stdout.write('.')
            sys.stdout.flush()
            err = e
        else:
            err = None
            break
        time.sleep(TRACE_MOUNT_TRY_DELAY)

    sys.stdout.write('\n')
    sys.stdout.flush()

    if err:
        raise e

def _is_mounted(attrs):
    try:
        simplecmd("mount | grep ' on %s type nfs'" % attrs.get('localpath'))
    except CmdErrorCode:
        return False
    return True

def mounts_exist():
    try:
        simplecmd("mount | grep ' type nfs'")
    except CmdErrorCode:
        return False
    return True

def _dir_exists():
    return os.path.exists(RUNNING_TRACE_DIR)

def _unmount(attrs):
    cmd('sudo umount %s' % attrs.get('localpath'))

def _try_unmount(attrs, quiet=False, cleanup=False):
    if not quiet:
        sys.stdout.write("Unmounting: %s..." % attrs.get('serverpath'))
        sys.stdout.flush()

    err = None

    tries = TRACE_UMOUNT_TRIES
    if cleanup:
        tries = TRACE_CLEANUP_UMOUNT_TRIES

    for i in range(tries):
        try:
            _unmount(attrs)
        except Exception, e:
            if not quiet:
                sys.stdout.write('.')
                sys.stdout.flush()
            err = e
        else:
            err = None
            break
        time.sleep(TRACE_UMOUNT_TRY_DELAY)

    if not quiet:
        sys.stdout.write('\n')

    if err:
        raise e

# TODO this should be bound to the parsing stuff!!
def _save_start_stats(attrs):
    commands = [ {'cmd': 'nfsstat',
                  'file': 'nfsstats.start'
                 },
                 {'cmd': 'dmesg',
                  'file': 'dmesg.start'
                 },
                 {'cmd': 'sudo sysctl -a | grep nfs',
                  'file': 'nfs_sysctls.start'
                 },
                 {'cmd': 'cat /proc/self/mountstats',
                  'file': 'proc_mountstats.start'
                 },
               ]

    _collect_stats(commands)

def _save_stop_stats(attrs):
    commands = [
        {'cmd': 'nfsstat',
         'file': 'nfsstats.stop'
        },
        {'cmd': 'nfsstat -S %s/nfsstats.start' % RUNNING_TRACE_DIR,
         'file': 'nfsstats'
        },
        {'cmd': 'dmesg',
         'file': 'dmesg.stop'
        },
        {'cmd': 'mountstats %s' % attrs.get('localpath'),
         'file': 'mountstats'
        },
        {'cmd': 'cat /proc/self/mountstats',
         'file': 'proc_mountstats.stop'
        },
        {'cmd': 'nfsiostat',
         'file': 'nfsiostat'
        },
        {'cmd': 'sudo sysctl -a | grep nfs',
         'file': 'nfs_sysctls.stop'
        },
    ]

    _collect_stats(commands)

def _collect_stats(commands):
    stats = []
    for c in commands:
        stats.append(c['file'])
        out = cmd(c['cmd'])
        f = file(os.path.join(RUNNING_TRACE_DIR, c['file']), 'w+')
        f.write('\n'.join(out[0]))

    #print 'Collecting stats: %s' % ', '.join(stats)


#
# public api commands
#
def start(mountopts, serverpath, workload, tags, is_setup=False):
    try:
        _start_x(mountopts, serverpath, workload, tags, is_setup=is_setup)
                 
    except Exception:
        print
        print 'Error starting trace, Cleaning up...'
        cleanup(serverpath)

        # reraise
        raise

def get_current_hostname():
    return simplecmd('hostname')

def get_current_kernel():
    return simplecmd('uname -r')

def _start_x(mountopts, serverpath, workload, tags, is_setup=False):

    if not tags:
        tags = []

    # gather any additional arguments
    hostname = get_current_hostname()
    kernel = get_current_kernel()

    #server, path = serverpath.split(':')
    m = _server_path_v6.match(serverpath)

    if not m:
        m = _server_path_v4.match(serverpath)

    if not m:
        raise ValueError("Cannot parse server, path from '%s'" % serverpath)

    server = m.group(1)
    path = m.group(2)

    _dir_create()

    attrs = TraceAttrs(new=True)

    attrs.set('mountopts', mountopts)
    attrs.set('serverpath', serverpath)
    attrs.set('server', server)
    attrs.set('path', path)
    attrs.set('localpath', MOUNTDIR)
    attrs.set('starttime', long(time.time()))
    attrs.set('workload', workload)
    attrs.set('kernel', get_current_kernel())
    attrs.set('client', get_current_hostname())
    attrs.set('tags', ','.join(tags))

    if is_setup:
        attrs.set('is_setup', 1)

    _try_mount(attrs)

    if not is_setup:
        _save_start_stats(attrs)

def stop(resdir=None):
    attrs = TraceAttrs(filename=os.path.join(RUNNING_TRACE_DIR, TRACE_ATTRFILE))

    if resdir != None and os.path.isdir(resdir):
        raise TraceError('Result directory already exists: %s' % resdir)

    attrs.set('stoptime', time.time())

    is_setup = long(attrs.get('is_setup', 0))

    if not is_setup:
        _save_stop_stats(attrs)

    if _is_mounted(attrs):
        _try_unmount(attrs)

    idle_check()

    if resdir != None:
        cmd('mv %s %s' % (RUNNING_TRACE_DIR, resdir))
        print 'Results copied to: %s' % (os.path.split(resdir)[-1],)
    else:
        cmd('rm -rf %s' % (RUNNING_TRACE_DIR))
        if not is_setup:
            print 'Results thrown away'

def cleanup(serverpath):
    attrs = TraceAttrs(temp=True)
    attrs.set('localpath', MOUNTDIR)
    attrs.set('serverpath', serverpath)

    if _is_mounted(attrs):
        _try_unmount(attrs, cleanup=True)

    if _dir_exists():
        _dir_aside()

def get_trace_attr(name):
    attrs = TraceAttrs(filename=os.path.join(RUNNING_TRACE_DIR, TRACE_ATTRFILE))
    return attrs.get(name)

#
#
#

def get_trace_list(resultdir, requested_workloads, mountopts_and_tags,
                   num_runs):
    # XXX fix this
    import collection

    # figure out what we already have run
    testcollection = collection.TestCollection(resultdir)


    #if not testcollection.is_empty():
    #    info = testcollection.load_info()
    #    print
    #    print "Current results directory contains %u %s:" % (info['total'], info['total_str'])
    #    print "  %s" % testcollection.load_info_str().replace('\n', '\n  ')

    if requested_workloads:
        workloads = {}

        for w in requested_workloads:
            try:
                workloads[w] = WORKLOADS[w]
            except KeyError:
                print
                warn('Invalid workload: "%s"' % w)
                print
                print "Available workloads:"
                print "  %s" % '\n  '.join(available_workloads())
                sys.exit(2)

    else:
        workloads = {}

        for w, workload_obj in WORKLOADS.iteritems():
            if not workload_obj.check():
                workloads[w] = workload_obj

    trace_list = []
    total = 0
    skipped = 0
    requested = 0

    current_kernel = get_current_kernel()

    for w, workload_obj in workloads.iteritems():
        for mountopt, tags in mountopts_and_tags:
            nfsopt = make_nfsopt(mountopt, tags)
            selection = collection.Selector(w, current_kernel, nfsopt)

            if testcollection.has_test(selection):
                testbin = testcollection.get_testbin(selection)
                already = testbin.num_runs()
            else:
                already = 0

            assert already >= 0

            need = num_runs - already
            if need < 0:
                need = 0

            if need > 0:
                trace_list.append((workload_obj, mountopt, tags, need))

            total += need
            requested += num_runs
            skipped += min(already, num_runs)

    return trace_list, workloads, total, requested, skipped

def _idle_check():
    # make sure there are no servers
    res = cmd('cat /proc/fs/nfsfs/servers 2>/dev/null | grep -v "^NV SERVER"',
              raiseerrorcode=False)

    res = '\n'.join(res[0]).strip()

    if res:
        raise Exception("NFS client not idle: %s" % res)

def idle_check():
    IDLE_MAX=2
    for i in range(IDLE_MAX):
        try:
            _idle_check()
        except:
            time.sleep(1)
            continue
        else:
            return

    _idle_check()

def run_traces(serverpath, tests_requested, resultdir, mountopts, num_runs,
               just_fetch, admin_script, admin_server, client_script,
               client_args, path_remote):

    mountopts = normalize_mountopts(mountopts)

    mountopts_and_tags = [ (m, None) for m in mountopts ]

    serveradmin = settings.ServerAdmin(admin_script, admin_server)

    if MOUNTOPTS_V4 in mountopts and \
       serveradmin.has_support(MOUNTOPTS_V4, TAG_DELEG_ENABLED):
        print "ServerAdmin v4.0 deleg support"
        mountopts_and_tags.append((MOUNTOPTS_V4, nfstags(TAG_DELEG_ENABLED)))

    if MOUNTOPTS_V4_1 in mountopts and \
       serveradmin.has_support(MOUNTOPTS_V4_1, TAG_PNFS_ENABLED):
        print "ServerAdmin v4.1 pnfs support"
        mountopts_and_tags.append((MOUNTOPTS_V4_1, nfstags(TAG_PNFS_ENABLED)))

    if path_remote:
        to_add = []
        for mo, ts in mountopts_and_tags:
            new_tags = []
            if ts:
                new_tags.extend(ts)
            new_tags.append(TAG_REMOTE)
            to_add.append((mo, nfstags(*new_tags)))

        mountopts_and_tags.extend(to_add)

    clientadmin = settings.ClientAdmin(client_script, client_args)

    client_tags = clientadmin.all_tag_combinations()

    if client_tags:
        new = []
        for mo, ts in mountopts_and_tags:
            for cts in client_tags:
                cts = cts.split(',')
                new_tags = []
                if ts:
                    new_tags.extend(ts)
                new_tags.extend(cts)
                new.append((mo, nfstags(*new_tags)))

        mountopts_and_tags = new

    trace_list, workloads, total, requested, skipped = \
        get_trace_list(resultdir, tests_requested, mountopts_and_tags,
                       num_runs)

    for w, workload_obj in workloads.iteritems():
        workload_obj.fetch()

    if just_fetch:
        sys.exit(0)

    # check each workload to make sure we'll be able to run it
    for w, workload_obj in workloads.iteritems():
        check_mesg = workload_obj.check()

        if check_mesg:
            raise ValueError("Workload %s is unavailable: %s" % (w, check_mesg))

    this_trace = 0

    print
    print "Requested: %u workloads X %u nfsopts X %u runs = %u test runs" % \
        (len(workloads), len(mountopts_and_tags), num_runs, requested)
    if skipped:
        print "Results directory already has %u matching traces" % (skipped,)
    print "Need to run %u of %u requested test runs" % (total, requested)

    for workload_obj, mountopt, tags, nruns in trace_list:
        print " %s - needs %u runs of %s " % \
            (workload_obj.name, nruns, make_nfsopt(mountopt, tags))
    print

    # cancel any ongoing trace
    cleanup(serverpath)

    dir_remove_old_asides()

    #if trace.mounts_exist():
        #mesg = 'This host already has NFS filesystems mounted.\n' + \
               #'Other NFS filesystems can mess with stats, and can kill ' + \
               #'running traces.'
        #usage(mesg)


    for workload_obj, mountopt, tags, nruns in trace_list:
        if nruns <= 0:
            continue

        serveradmin.handle_settings(mountopt, tags)
        clientadmin.handle_settings(mountopt, tags)

        this_serverpath = serverpath
        if tags and TAG_REMOTE in tags:
            this_serverpath = "%s:%s" % \
                (serverpath.split(':', 1)[0], path_remote)

        for run in range(nruns):
            this_trace += 1

            inform("Trace %u/%u: " % (this_trace, total) + \
                "%u/%u with workload: %s, nfsopts: %s" % \
                (run+1, nruns, workload_obj.name, make_nfsopt(mountopt, tags)))

            # check server health
            serveradmin.health()

            highlight("< SETUP >")
            start(mountopt, this_serverpath, workload_obj.name, tags,
                  is_setup=True)

            try:
                workload_obj.setup()
            except Exception, e:
                warn("An error occurred!  Cleaning up...")

                cleanup(serverpath)
                raise e # re-raise exception

            except KeyboardInterrupt:
                warn("An error occurred!  Cleaning up...")

                cleanup(serverpath)
                raise KeyboardInterrupt

            stop()

            highlight("< RUN >")
            start(mountopt, this_serverpath, workload_obj.name, tags)

            try:
                workload_obj.run()

            except Exception, e:
                warn("An error occurred!  Cleaning up...")

                cleanup(serverpath)
                raise e # re-raise exception

            except KeyboardInterrupt:
                warn("An error occurred!  Cleaning up...")

                cleanup(serverpath)
                raise KeyboardInterrupt

            tracedir = get_tracedir(resultdir, workload_obj.name,
                                    make_nfsopt(mountopt, tags))


            stop(os.path.join(resultdir, tracedir))

            try:
                # check server health
                serveradmin.health()
            except:
                cmd('rm -rf %s' % tracedir)
                # reraise
                raise

            print

    if this_trace == 0:
        inform('No workload runs were needed!')
    else:
        inform('Successfully ran %u traces!' % (this_trace,))

